@@ -39,7 +39,11 @@ void manager_reset_config(Manager *m) {
         m->n_autovts = 6;
         m->reserve_vt = 6;
 #endif // 0
+#if 0 /// elogind must not remove IPC items by default (See https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=949698)
         m->remove_ipc = true;
+#else // 0
+        m->remove_ipc = false;
+#endif // 0
         m->inhibit_delay_max = 5 * USEC_PER_SEC;
 #if 0 /// elogind does not start a user service manager, the delay is unneeded.
         m->user_stop_delay = 10 * USEC_PER_SEC;
@@ -389,10 +393,8 @@ int manager_get_session_by_pidref(Manager *m, const PidRef *pid, Session **ret)
         } else {
 #if 0 /// elogind does not support systemd units, but its own session system
                 r = cg_pidref_get_unit(pid, &unit);
-                if (r < 0)
-                        return r;
-
-                s = hashmap_get(m->session_units, unit);
+                if (r >= 0)
+                        s = hashmap_get(m->session_units, unit);
 #else // 0
                 log_debug_elogind("Searching session for PID %d", pid->pid);
                 r = cg_pid_get_session(pid->pid, &session_name);
@@ -446,10 +448,14 @@ int manager_get_user_by_pid(Manager *m, pid_t pid, User **ret) {
 int manager_get_idle_hint(Manager *m, dual_timestamp *t) {
         Session *s;
         bool idle_hint;
-        dual_timestamp ts = DUAL_TIMESTAMP_NULL;
+        dual_timestamp ts;
 
         assert(m);
 
+        /* Initialize the baseline timestamp with the time the manager got initialized to avoid reporting
+         * unreasonable large idle periods starting with the Unix epoch. */
+        ts = m->init_ts;
+
         idle_hint = !manager_is_inhibited(m, INHIBIT_IDLE, INHIBIT_BLOCK, t, false, false, 0, NULL);
 
         HASHMAP_FOREACH(s, m->sessions) {
