@@ -89,12 +89,6 @@ static int manager_new(Manager **ret) {
 #endif // 0
                 return -ENOMEM;
 
-#if 1 /// elogind needs some more data
-        r = elogind_manager_new(m);
-        if (r < 0) {
-                return r;
-        }
-#endif // 1
         r = sd_event_default(&m->event);
         if (r < 0)
                 return r;
@@ -119,6 +113,15 @@ static int manager_new(Manager **ret) {
 
         (void) sd_event_set_watchdog(m->event, true);
 
+        dual_timestamp_now(&m->init_ts);
+
+#if 1 /// elogind needs some more data
+        r = elogind_manager_new(m);
+        if (r < 0) {
+                return r;
+        }
+#endif // 1
+
         manager_reset_config(m);
 
         *ret = TAKE_PTR(m);
@@ -196,9 +199,7 @@ static Manager* manager_free(Manager *m) {
 
         free(m->scheduled_shutdown_tty);
         free(m->wall_message);
-#if 0 /// UNNEEDED by elogind
         free(m->action_job);
-#endif // 0
 
         strv_free(m->efi_boot_loader_entries);
         free(m->efi_loader_entry_one_shot);
@@ -958,6 +959,7 @@ static void manager_gc(Manager *m, bool drop_not_started) {
                 seat->in_gc_queue = false;
 
                 if (seat_may_gc(seat, drop_not_started)) {
+                        log_debug_elogind("Seat '%s' from gc queue will be dropped now", strna(seat->id));
                         seat_stop(seat, /* force = */ false);
                         seat_free(seat);
                 }
@@ -974,6 +976,7 @@ static void manager_gc(Manager *m, bool drop_not_started) {
                 /* Normally, this should make the session referenced again, if it doesn't then let's get rid
                  * of it immediately. */
                 if (session_may_gc(session, drop_not_started)) {
+                        log_debug_elogind("Session '%s' from gc queue will be finalized now", strna(session->id));
                         (void) session_finalize(session);
                         session_free(session);
                 }
@@ -988,6 +991,7 @@ static void manager_gc(Manager *m, bool drop_not_started) {
 
                 /* Second step: finalize user */
                 if (user_may_gc(user, drop_not_started)) {
+                        log_debug_elogind("User '%s' from gc queue will be finalized now", strna(user->user_record->user_name));
                         (void) user_finalize(user);
                         user_free(user);
                 }
@@ -1308,6 +1312,7 @@ static int run(int argc, char *argv[]) {
 #endif // 0
 
         log_debug_elogind("%s", "Creating manager...");
+
 #if 0 /// elogind also blocks SIGQUIT, and installs a signal handler for it
         assert_se(sigprocmask_many(SIG_BLOCK, NULL, SIGHUP, SIGTERM, SIGINT, SIGCHLD, SIGRTMIN+18, -1) >= 0);
 #else // 0
