@@ -158,6 +158,18 @@ static int do_execute(
                 }
 
 #if 0 /// elogind never sets this PID as it is never INIT
+                if (FLAGS_SET(flags, EXEC_DIR_WARN_WORLD_WRITABLE)) {
+                        struct stat st;
+
+                        r = stat(t, &st);
+                        if (r < 0)
+                                log_warning_errno(errno, "Failed to stat '%s', ignoring: %m", t);
+                        else if (S_ISREG(st.st_mode) && (st.st_mode & 0002))
+                                log_warning("'%s' is marked world-writable, which is a security risk as it "
+                                            "is executed with privileges. Please remove world writability "
+                                            "permission bits. Proceeding anyway.", t);
+                }
+
                 r = do_spawn(t, argv, fd, &pid, FLAGS_SET(flags, EXEC_DIR_SET_SYSTEMD_EXEC_PID));
 #else // 0
                 r = do_spawn(t, argv, fd, &pid);
@@ -193,8 +205,25 @@ static int do_execute(
                                         return log_error_errno(errno, "Failed to seek on serialization fd: %m");
 
                                 r = callbacks[STDOUT_GENERATE](TAKE_FD(fd), callback_args[STDOUT_GENERATE]);
+#if 0 /// elogind allows scripts to cancel the running operation
                                 if (r < 0)
                                         return log_error_errno(r, "Failed to process output from %s: %m", *path);
+#else // 0
+                                if (r) {
+                                        if (ECANCELED == r) {
+                                                /* The operation was cancelled by the script, but elogind is
+                                                 * configured to proceed anyway.
+                                                 * gather_output_generate() already logged the scripts output. */
+                                                log_warning_errno(r, "%s: %m", *path);
+                                        } else if (-ECANCELED == r) {
+                                                /* The operation was cancelled by the script, and elogind is
+                                                 * configured to break off of that happens. */
+                                                log_error_errno(r, "%s: %m", *path);
+                                                skip_remaining = true;
+                                        } else
+                                                return log_error_errno(r, "Failed to process output from %s: %m", *path);
+                                }
+#endif // 0
                         }
 
                         if (skip_remaining)
@@ -617,5 +646,6 @@ int fork_agent(const char *name, const int except[], size_t n_except, pid_t *ret
         va_end(ap);
 
         execv(path, l);
+        log_error_errno(errno, "Failed to execute %s: %m", path);
         _exit(EXIT_FAILURE);
 }
